/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

void alloc_log_likelihood(void)
{
  log_likelihood = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);
  lambda = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_smooth)
    log_prior = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);

  log_likelihood_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);
  lambda_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_smooth)
    log_prior_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);
}

void alloc_log_itemp(void)
{
  log_itemp = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);
  log_itemp_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*log_likelihood_size);
}

void expand_log_likelihood_all(void)
{
  int old_log_likelihood_size;

  old_log_likelihood_size = log_likelihood_size;
  log_likelihood_size *= 2;

  log_likelihood = (FLOAT_T *)n_realloc(log_likelihood,sizeof(FLOAT_T)*log_likelihood_size);
  lambda = (FLOAT_T *)n_realloc(lambda,sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_smooth)
    log_prior = (FLOAT_T *)n_realloc(log_prior,sizeof(FLOAT_T)*log_likelihood_size);

  log_likelihood_best = (FLOAT_T *)n_realloc(log_likelihood_best,sizeof(FLOAT_T)*log_likelihood_size);
  lambda_best = (FLOAT_T *)n_realloc(lambda_best,sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_smooth)
    log_prior_best = (FLOAT_T *)n_realloc(log_prior_best,sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_daem) {
    log_itemp = (FLOAT_T *)n_realloc(log_itemp,sizeof(FLOAT_T)*log_likelihood_size);
    log_itemp_best = (FLOAT_T *)n_realloc(log_itemp_best,sizeof(FLOAT_T)*log_likelihood_size);
  }
}

void alloc_prob(void)
{
  int k;

  prob_comp = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
  prob_cond = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++) {
    prob_comp[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
    prob_cond[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
  }

  prob_incomp = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
}

void alloc_ic(int num_search)
{
  ic = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_search);
  bic = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_search);
  autoclass = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_search);
}

void alloc_seed(void)
{
  int u,num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  log_seed = (unsigned int **)n_malloc(sizeof(unsigned int *)*num_search);
  log_num_iterate = (int **)n_malloc(sizeof(int *)*num_search);
  log_final_lambda = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_search);
  log_r_best = (int *)n_malloc(sizeof(int)*num_search);

  for (u = 0; u < num_search; u++) {
    log_seed[u] = (unsigned int *)n_malloc(sizeof(unsigned int)*num_init);
    log_num_iterate[u] = (int *)n_malloc(sizeof(int)*num_init);
    log_final_lambda[u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_init);
  }
}

void release_prob(void)
{
  int k;
  for (k=0; k<num_class; k++) {
    free(prob_comp[k]); prob_comp[k] = NULL;
    free(prob_cond[k]); prob_cond[k] = NULL;
  }
  free(prob_comp);   prob_comp = NULL;
  free(prob_incomp); prob_incomp = NULL;
  free(prob_cond);   prob_cond = NULL;
}

void release_log_likelihood(void)
{
  free(log_likelihood); log_likelihood = NULL;
  free(lambda); lambda = NULL;
  if (mode_smooth) {
    free(log_prior); log_prior = NULL;
  }

  free(log_likelihood_best); log_likelihood_best = NULL;
  free(lambda_best); lambda_best = NULL;
  if (mode_smooth) {
    free(log_prior_best); log_prior_best = NULL;
  }
}

void release_log_itemp(void)
{
  free(log_itemp); log_itemp = NULL;
  free(log_itemp_best); log_itemp_best = NULL;
}

void release_ic(void)
{
  free(ic);  ic = NULL;
  free(bic); bic = NULL;
  free(autoclass); autoclass = NULL;
}

void release_seed(void)
{
  int u, num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  for (u=0; u < num_search; u++) {
    free(log_seed[u]);         log_seed[u] = NULL;
    free(log_num_iterate[u]);  log_num_iterate[u] = NULL;
    free(log_final_lambda[u]); log_final_lambda[u] = NULL;
  }
  free(log_seed);         log_seed = NULL;
  free(log_num_iterate);  log_num_iterate = NULL;
  free(log_final_lambda); log_final_lambda = NULL;
}

void alloc_param(void)
{
  int k,v,u,n;

  param_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    for (k=0; k<num_class; k++) {
      param_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	param_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    param_u = (VECTOR **)n_malloc(sizeof(MATRIX *)*num_class);
    param_s = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    param_p = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    param_sdet = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      param_s[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      param_p[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      param_sdet[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	param_u[k][u] = alloc_vector(n);
	param_s[k][u] = alloc_matrix(n,n);
	param_p[k][u] = alloc_matrix(n,n);
      }
    }
  }

}

void alloc_param_best(void)
{
  int k,v,u,n;

  param_c_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w_best = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      param_w_best[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	param_w_best[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }
    
  if (mode_gauss) {
    param_u_best = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    param_s_best = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_best[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      param_s_best[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	param_u_best[k][u] = alloc_vector(n);
	param_s_best[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void save_param_best(void)
{
  int k,v,j,u;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c_best[k] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
	for (j=0; j<num_word[v]; j++)
	  param_w_best[k][v][j] = param_w[k][v][j];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(param_u_best[k][u],param_u[k][u]);
	copy_matrix(param_s_best[k][u],param_s[k][u]);
      }
    }
  }
}

void restore_param_best(void)
{
  int k,v,j,u;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    param_c[k] = param_c_best[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
	for (j=0; j<num_word[v]; j++)
	  param_w[k][v][j] = param_w_best[k][v][j];
  }

  if (mode_gauss) {
#pragma omp parallel for private(u)
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(param_u[k][u],param_u_best[k][u]);
	copy_matrix(param_s[k][u],param_s_best[k][u]);
	param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
	if (err_inv) die("Error in inverting covariances\n");
      }
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void save_param_opt(void)
{
  int k,v,i,u,n;

  param_c_opt = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class_opt);

  if (mode_nominal) {
    param_w_opt = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class_opt);
    
    for (k=0; k<num_class_opt; k++) {
      param_w_opt[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	param_w_opt[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class_opt; k++)
    param_c_opt[k] = param_c[k];

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class_opt; k++)
      for (v=0; v<tuple_size_d; v++)
	for (i=0; i<num_word[v]; i++)
	  param_w_opt[k][v][i] = param_w[k][v][i];
  }

  if (mode_gauss) {
    param_u_opt = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    param_s_opt = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_opt[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      param_s_opt[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);

      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	param_u_opt[k][u] = alloc_vector(n);
	param_s_opt[k][u] = alloc_matrix(n,n);
      }
    }

#pragma omp parallel for private(u)
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(param_u_opt[k][u],param_u[k][u]);
	copy_matrix(param_s_opt[k][u],param_s[k][u]);
      }
    }
  }
}

/*
 * Since the number of parameters can vary, we should allocate
 * the new memory area (note that the size is based on num_class_opt).
 */
void restore_param_opt(void)
{
  int k,v,i,u,n;

  param_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class_opt);

  if (mode_nominal) {
    param_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class_opt);
    
    for (k=0; k<num_class_opt; k++) {
      param_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	param_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class_opt; k++)
    param_c[k] = param_c_opt[k];

  if (mode_nominal) {
#pragma omp parallel for private(i,v)
    for (k=0; k<num_class_opt; k++)
      for (v=0; v<tuple_size_d; v++)
	for (i=0; i<num_word[v]; i++)
	  param_w[k][v][i] = param_w_opt[k][v][i];
  }

  if (mode_gauss) {
    param_u = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class_opt);
    param_s = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class_opt);
    param_p = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class_opt);
    param_sdet = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class_opt);

    for (k=0; k<num_class_opt; k++) {
      param_u[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      param_s[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      param_p[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      param_sdet[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
        param_u[k][u] = alloc_vector(n);
        param_s[k][u] = alloc_matrix(n,n);
        param_p[k][u] = alloc_matrix(n,n);
      }
    }

#pragma omp parallel for private(u)
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(param_u[k][u],param_u_opt[k][u]);
	copy_matrix(param_s[k][u],param_s_opt[k][u]);
	param_sdet[k][u] = INVERT(param_s[k][u],param_p[k][u]);
	if (err_inv) die("Error in inverting covariances\n");
      }
    }
  }

  if (mode_gauss) alloc_hparam_gauss();
}

void alloc_hparam(void)
{
  int k,v;

  hparam_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++)
	hparam_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }
}

void release_hparam(void)
{
  int k,v;

  free(hparam_c); hparam_c = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(hparam_w[k][v]); hparam_w[k][v] = NULL;
      }
      free(hparam_w[k]); hparam_w[k] = NULL;
    }
    free(hparam_w); hparam_w = NULL;
  }    
}

void alloc_hparam_gauss(void)
{
  int k,u,n;

/* A bit dirty.. */
#ifdef VBNC
  if (mode_gauss && mode_smooth_gauss) {
#else
  if (mode_gauss) {
#endif

    hparam_u = (VECTOR **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_s = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
    
    for (k=0; k<num_class; k++) {
      hparam_u[k] = (VECTOR *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_s[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	hparam_u[k][u] = alloc_vector(n);
	hparam_s[k][u] = alloc_matrix(n,n);
	hparam_p[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void release_hparam_gauss(void)
{
  int k,u;

  if (mode_gauss && mode_smooth_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(hparam_u[k][u]);
	release_matrix(hparam_s[k][u]);
	release_matrix(hparam_p[k][u]);
      }
      free(hparam_u[k]);
      free(hparam_s[k]);
      free(hparam_p[k]);
      free(hparam_sdet[k]);
    }
    free(hparam_u);  hparam_u = NULL;
    free(hparam_s);  hparam_s = NULL;
    free(hparam_p);  hparam_p = NULL;
    free(hparam_sdet);  hparam_sdet = NULL;
  }
}

void alloc_hparam_gauss_aux(void)
{
  int k;

  hparam_tau = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
  hparam_alpha_w = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

  for (k=0; k<num_class; k++)
    hparam_alpha_w[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
}

void release_hparam_gauss_aux(void)
{
  int k;

  for (k=0; k<num_class; k++)
    free(hparam_alpha_w[k]);

  free(hparam_tau);      hparam_tau = NULL;
  free(hparam_alpha_w);  hparam_alpha_w = NULL;
}

void alloc_expect(void)
{
  int k,v,u,n;

  expect_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    expect_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      expect_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	expect_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    expect_u = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    expect_s = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);

    for (k=0; k<num_class; k++) {
      expect_u[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      expect_s[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);

      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	expect_u[k][u] = alloc_vector(n);
	expect_s[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void alloc_prob_mem(void)
{
  int k,v,u;

  prob_mem_case = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++)
    prob_mem_case[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);

  if (mode_nominal) {
    prob_mem_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      prob_mem_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	prob_mem_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
    
    prob_w = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      prob_w[v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
  }

  if (mode_gauss) {
    prob_mem_g = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      prob_mem_g[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
        prob_mem_g[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
    }

    prob_g = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      prob_g[u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
  }
}

void release_param(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(param_w[k][v]); param_w[k][v] = NULL;
      }
      free(param_w[k]); param_w[k] = NULL;
    }
    free(param_w); param_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
        release_vector(param_u[k][u]);
        release_matrix(param_s[k][u]);
        release_matrix(param_p[k][u]);
      }
      free(param_u[k]);
      free(param_s[k]);
      free(param_p[k]);
      free(param_sdet[k]);
    }
    free(param_u);     param_u = NULL;
    free(param_s);     param_s = NULL;
    free(param_p);     param_p = NULL;
    free(param_sdet);  param_sdet = NULL;
  }

  free(param_c); param_c = NULL;
}

void release_param_best(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(param_w_best[k][v]); param_w_best[k][v] = NULL;
      }
      free(param_w_best[k]); param_w_best[k] = NULL;
    }
    free(param_w_best); param_w_best = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(param_u_best[k][u]);
	release_matrix(param_s_best[k][u]);
      }
      free(param_u_best[k]);
      free(param_s_best[k]);
    }
    free(param_u_best); param_u_best = NULL;
    free(param_s_best); param_s_best = NULL;
  }

  free(param_c_best); param_c_best = NULL;
}

void release_param_opt(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class_opt; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(param_w_opt[k][v]); param_w_opt[k][v] = NULL;
      }
      free(param_w_opt[k]); param_w_opt[k] = NULL;
    }
    free(param_w_opt); param_w_opt = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(param_u_opt[k][u]);
	release_matrix(param_s_opt[k][u]);
      }
      free(param_u_opt[k]);
      free(param_s_opt[k]);
    }
    free(param_u_opt); param_u_opt = NULL;
    free(param_s_opt); param_s_opt = NULL;
  }

  free(param_c_opt); param_c_opt = NULL;
}

void release_expect(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(expect_w[k][v]); expect_w[k][v] = NULL;
      }
      free(expect_w[k]); expect_w[k] = NULL;
    }
    free(expect_w); expect_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(expect_u[k][u]);
	release_matrix(expect_s[k][u]);
      }
      free(expect_u[k]);
      free(expect_s[k]);
    }
    free(expect_u); expect_u = NULL;
    free(expect_s); expect_s = NULL;
  }

  free(expect_c); expect_c = NULL;
}

void release_prob_mem(void)
{
  int k,v,u;

  for (k=0; k<num_class; k++)
    free(prob_mem_case[k]);
  free(prob_mem_case); prob_mem_case = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
	free(prob_mem_w[k][v]);
      free(prob_mem_w[k]);
    }
    free(prob_mem_w); prob_mem_w = NULL;

    for (v=0; v<tuple_size_d; v++)
      free(prob_w[v]);
    free(prob_w); prob_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++)
        free(prob_mem_g[k][u]); 
      free(prob_mem_g[k]);
    }
    free(prob_mem_g); prob_mem_g = NULL;

    for (u=0; u<gauss_size; u++)
      free(prob_g[u]);
    free(prob_g); prob_g = NULL;
  }
}

void alloc_read_param(int max_num_class, int *max_num_word)
{
  int k,v,u,a;

  read_param_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*max_num_class);

  if (mode_nominal) {
    read_param_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*max_num_class);
    
    for (k=0; k<max_num_class; k++) {
      read_param_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	read_param_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*max_num_word[v]);
    }
  }

  if (mode_gauss) {
    read_param_u = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*max_num_class);
    read_param_s = (FLOAT_T ****)n_malloc(sizeof(FLOAT_T ***)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_param_u[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      read_param_s[k] = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	read_param_u[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
	read_param_s[k][u] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
	
	for (a=0; a<gauss_attr_size[u]; a++)
	  read_param_s[k][u][a] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
  }
}

void release_read_param(int max_num_class)
{
  int k,v,u,a;

  free(read_param_c); read_param_c = NULL;

  if (mode_nominal) {
    for (k=0; k<max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(read_param_w[k][v]); read_param_w[k][v] = NULL;
      }
      free(read_param_w[k]); read_param_w[k] = NULL;
    }
    free(read_param_w); read_param_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	free(read_param_u[k][u]); 
	for (a=0; a<gauss_attr_size[u]; a++) {
	  free(read_param_s[k][u][a]);
	}
      }
      free(read_param_u[k]);
      free(read_param_s[k]);
    }
    free(read_param_u); read_param_u = NULL;
    free(read_param_s); read_param_s = NULL;
  }
}

void expand_read_param_c(int old_max_num_class, int new_max_num_class)
{
  int i;
  FLOAT_T *new_read_param_c;

  new_read_param_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*new_max_num_class);

  for (i=0; i<old_max_num_class; i++)
    new_read_param_c[i] = read_param_c[i];

  free(read_param_c);
  read_param_c = new_read_param_c;
}

void expand_read_param_w(int old_max_num_class, int new_max_num_class,
			 int *old_max_num_word, int *new_max_num_word)
{
  int k,i,v;
  FLOAT_T ***new_read_param_w;

  if (mode_nominal) {
    new_read_param_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_param_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      
      for (v=0; v<tuple_size_d; v++)
	new_read_param_w[k][v] =
	  (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*new_max_num_word[v]);
    }
    
    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (i=0; i<old_max_num_word[v]; i++) {
	  new_read_param_w[k][v][i] = read_param_w[k][v][i];
	}
      }
    }
    
    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(read_param_w[k][v]); read_param_w[k][v] = NULL;
      }
      free(read_param_w[k]); read_param_w[k] = NULL;
    }
    free(read_param_w);
    
    read_param_w = new_read_param_w;
  }
}

void expand_read_param_u(int old_max_num_class, int new_max_num_class)
{
  int k,u,a;
  FLOAT_T ***new_read_param_u;

  if (mode_gauss) {
    new_read_param_u = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_param_u[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
	new_read_param_u[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
    }
    
    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
	for (a=0; a<gauss_attr_size[u]; a++)
	  new_read_param_u[k][u][a] = read_param_u[k][u][a];
    
    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	free(read_param_u[k][u]);
      }
      free(read_param_u[k]);
    }
    free(read_param_u);
    
    read_param_u = new_read_param_u;
  }
}

void expand_read_param_s(int old_max_num_class, int new_max_num_class)
{
  int k,u,a,a0,a1;
  FLOAT_T ****new_read_param_s;
  
  if (mode_gauss) {
    new_read_param_s = (FLOAT_T ****)n_malloc(sizeof(FLOAT_T ***)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_param_s[k] = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	new_read_param_s[k][u] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
	for (a=0; a<gauss_attr_size[u]; a++)
	  new_read_param_s[k][u][a] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
    
    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
	for (a0=0; a0<gauss_attr_size[u]; a0++)
	  for (a1=0; a1<gauss_attr_size[u]; a1++)
	    new_read_param_s[k][u][a0][a1] = read_param_s[k][u][a0][a1];
    
    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	for (a=0; a<gauss_attr_size[u]; a++) {
	  free(read_param_s[k][u][a]);
	}
	free(read_param_s[k][u]);
      }
      free(read_param_s[k]);
    }
    free(read_param_s);
    
    read_param_s = new_read_param_s;
  }
}

void alloc_read_hparam(int max_num_class, int *max_num_word)
{
  int k,v,u,a;

  read_hparam_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*max_num_class);

  if (mode_nominal) {
    read_hparam_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*max_num_class);
    
    for (k=0; k<max_num_class; k++) {
      read_hparam_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	read_hparam_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*max_num_word[v]);
    }
  }

  if (mode_gauss) {
    read_hparam_u = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*max_num_class);
    read_hparam_s = (FLOAT_T ****)n_malloc(sizeof(FLOAT_T ***)*max_num_class);

    for (k=0; k<max_num_class; k++) {
      read_hparam_u[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      read_hparam_s[k] = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	read_hparam_u[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
	read_hparam_s[k][u] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
	
	for (a=0; a<gauss_attr_size[u]; a++)
	  read_hparam_s[k][u][a] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }

    read_tau = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*max_num_class);
    read_alpha_w = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*max_num_class);
    for (k=0; k<max_num_class; k++)
      read_alpha_w[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
  }
}

void release_read_hparam(int max_num_class)
{
  int k,v,u,a;

  free(read_hparam_c); read_hparam_c = NULL;

  if (mode_nominal) {
    for (k=0; k<max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(read_hparam_w[k][v]); read_hparam_w[k][v] = NULL;
      }
      free(read_hparam_w[k]); read_hparam_w[k] = NULL;
    }
    free(read_hparam_w); read_hparam_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	free(read_hparam_u[k][u]); 
	for (a=0; a<gauss_attr_size[u]; a++) {
	  free(read_hparam_s[k][u][a]);
	}
      }
      free(read_hparam_u[k]);
      free(read_hparam_s[k]);
    }
    free(read_hparam_u); read_hparam_u = NULL;
    free(read_hparam_s); read_hparam_s = NULL;
 
    for (k=0; k<num_class; k++) {
      free(read_alpha_w[k]);
    }

    free(read_tau);     read_tau = NULL;
    free(read_alpha_w); read_alpha_w = NULL;
}
}

void expand_read_hparam_c(int old_max_num_class, int new_max_num_class)
{
  int i;
  FLOAT_T *new_read_hparam_c;

  new_read_hparam_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*new_max_num_class);

  for (i=0; i<old_max_num_class; i++)
    new_read_hparam_c[i] = read_hparam_c[i];

  free(read_hparam_c);
  read_hparam_c = new_read_hparam_c;
}

void expand_read_hparam_w(int old_max_num_class, int new_max_num_class,
			 int *old_max_num_word, int *new_max_num_word)
{
  int k,i,v;
  FLOAT_T ***new_read_hparam_w;

  if (mode_nominal) {
    new_read_hparam_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      
      for (v=0; v<tuple_size_d; v++)
	new_read_hparam_w[k][v] =
	  (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*new_max_num_word[v]);
    }
    
    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (i=0; i<old_max_num_word[v]; i++) {
	  new_read_hparam_w[k][v][i] = read_hparam_w[k][v][i];
	}
      }
    }
    
    for (k=0; k<old_max_num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(read_hparam_w[k][v]); read_hparam_w[k][v] = NULL;
      }
      free(read_hparam_w[k]); read_hparam_w[k] = NULL;
    }
    free(read_hparam_w);
    
    read_hparam_w = new_read_hparam_w;
  }
}

void expand_read_hparam_u(int old_max_num_class, int new_max_num_class)
{
  int k,u,a;
  FLOAT_T ***new_read_hparam_u;

  if (mode_gauss) {
    new_read_hparam_u = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_u[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      for (u=0; u<gauss_size; u++)
	new_read_hparam_u[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
    }
    
    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
	for (a=0; a<gauss_attr_size[u]; a++)
	  new_read_hparam_u[k][u][a] = read_hparam_u[k][u][a];
    
    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	free(read_hparam_u[k][u]);
      }
      free(read_hparam_u[k]);
    }
    free(read_hparam_u);
    
    read_hparam_u = new_read_hparam_u;
  }
}

void expand_read_hparam_s(int old_max_num_class, int new_max_num_class)
{
  int k,u,a,a0,a1;
  FLOAT_T ****new_read_hparam_s;
  
  if (mode_gauss) {
    new_read_hparam_s = (FLOAT_T ****)n_malloc(sizeof(FLOAT_T ***)*new_max_num_class);
    
    for (k=0; k<new_max_num_class; k++) {
      new_read_hparam_s[k] = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	new_read_hparam_s[k][u] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);
	for (a=0; a<gauss_attr_size[u]; a++)
	  new_read_hparam_s[k][u][a] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
    
    for (k=0; k<old_max_num_class; k++)
      for (u=0; u<gauss_size; u++)
	for (a0=0; a0<gauss_attr_size[u]; a0++)
	  for (a1=0; a1<gauss_attr_size[u]; a1++)
	    new_read_hparam_s[k][u][a0][a1] = read_hparam_s[k][u][a0][a1];
    
    for (k=0; k<old_max_num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	for (a=0; a<gauss_attr_size[u]; a++) {
	  free(read_hparam_s[k][u][a]);
	}
	free(read_hparam_s[k][u]);
      }
      free(read_hparam_s[k]);
    }
    free(read_hparam_s);
    
    read_hparam_s = new_read_hparam_s;
  }
}

void alloc_cluster(void)
{
  int k,v,u;

  if (mode_nominal) {
    cluster_w_pred = (int **)n_malloc(sizeof(int *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      cluster_w_pred[v] = (int *)n_malloc(sizeof(int)*num_word[v]);
    
    cluster_w_freq = (int **)n_malloc(sizeof(int *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      cluster_w_freq[v] = (int *)n_malloc(sizeof(int)*num_class);
    
    cluster_w = (int ***)n_malloc(sizeof(int **)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++) {
      cluster_w[v] = (int **)n_malloc(sizeof(int *)*num_class);
      for (k=0; k<num_class; k++)
	cluster_w[v][k] = (int *)n_malloc(sizeof(int)*num_word[v]);
    }
  }
    
  if (mode_gauss) {
    cluster_g_pred = (int **)n_malloc(sizeof(int *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      cluster_g_pred[u] = (int *)n_malloc(sizeof(int)*num_entry);
    
    cluster_g_freq = (int **)n_malloc(sizeof(int *)*gauss_size);
    for (u=0; u<gauss_size; u++)
      cluster_g_freq[u] = (int *)n_malloc(sizeof(int)*num_class);
    
    cluster_g = (int ***)n_malloc(sizeof(int **)*gauss_size);
    for (u=0; u<gauss_size; u++) {
      cluster_g[u] = (int **)n_malloc(sizeof(int *)*num_class);
      for (k=0; k<num_class; k++)
	cluster_g[u][k] = (int *)n_malloc(sizeof(int)*num_entry);
    }
  }

  cluster_case_pred = (int *)n_malloc(sizeof(int)*num_entry);  /* redundant, but safe */
  cluster_case_freq = (int *)n_malloc(sizeof(int)*num_class);

  cluster_case = (int **)n_malloc(sizeof(int *)*num_class);
  for (k=0; k<num_class; k++)
    cluster_case[k] = (int *)n_malloc(sizeof(int)*num_entry);  /* redundant, but safe */
}

void release_cluster(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++) {
      for (k=0; k<num_class; k++)
	free(cluster_w[v][k]);
      free(cluster_w[v]);
    }
    free(cluster_w); cluster_w = NULL;
    
    for (v=0; v<tuple_size_d; v++) {
      free(cluster_w_pred[v]);
      free(cluster_w_freq[v]);
    }
    free(cluster_w_pred); cluster_w_pred = NULL;
    free(cluster_w_freq); cluster_w_freq = NULL;
  }
    
  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      for (k=0; k<num_class; k++)
	free(cluster_g[u][k]);
      free(cluster_g[u]);
    }
    free(cluster_g); cluster_g = NULL;
    
    for (u=0; u<gauss_size; u++) {
      free(cluster_g_pred[u]);
      free(cluster_g_freq[u]);
    }
    free(cluster_g_pred); cluster_g_pred = NULL;
    free(cluster_g_freq); cluster_g_freq = NULL;
  }

  for (k=0; k<num_class; k++) {
    free(cluster_case[k]); cluster_case[k] = NULL;
  }
  free(cluster_case); cluster_case = NULL;
  
  free(cluster_case_pred); cluster_case_pred = NULL;
  free(cluster_case_freq); cluster_case_freq = NULL;
}

void alloc_rank(void)
{
  int k,v,u;

  rank_class = (int *)n_malloc(sizeof(int)*num_class);
  rank_mem_case = (int **)n_malloc(sizeof(int **)*num_class);

  for (k=0; k<num_class; k++)
    rank_mem_case[k] = (int *)n_malloc(sizeof(int)*num_entry);

  if (mode_nominal) {
    rank_mem_w = (int ***)n_malloc(sizeof(int **)*num_class);
    
    for (k=0; k<num_class; k++) {
      rank_mem_w[k] = (int **)n_malloc(sizeof(int *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	rank_mem_w[k][v] = (int *)n_malloc(sizeof(int)*num_word[v]);
    }
  }

  if (mode_gauss) {
    rank_mem_g = (int ***)n_malloc(sizeof(int **)*num_class);
    
    for (k=0; k<num_class; k++) {
      rank_mem_g[k] = (int **)n_malloc(sizeof(int *)*gauss_size);
      for (u=0; u<gauss_size; u++)
	rank_mem_g[k][u] = (int *)n_malloc(sizeof(int)*num_entry);
    }
  }
}

void release_rank(void)
{
  int k,v,u;

  free(rank_class); rank_class = NULL;

  for (k=0; k<num_class; k++) {
    free(rank_mem_case[k]); rank_mem_case[k] = NULL;
  }
  free(rank_mem_case); rank_mem_case = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++)
	free(rank_mem_w[k][v]);
      free(rank_mem_w[k]);
    }
    free(rank_mem_w); rank_mem_w = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++)
	free(rank_mem_g[k][u]);
      free(rank_mem_g[k]);
    }
    free(rank_mem_g); rank_mem_g = NULL;
  }
}

void alloc_log_likelihood_sm_size(void)
{
  int t;

  log_likelihood_sm_size = (int *)n_malloc(sizeof(int)*(num_sm + 1));

  for (t = 0; t < (num_sm + 1); t++) {
    log_likelihood_sm_size[t] = MAX_MAX_ITERATE;
  }
}

void alloc_log_likelihood_sm(void)
{
  int n,t;

  /* We need one extra entry to record the values for the first full EM */
  log_likelihood_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm+1));
  log_prior_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm+1));
  lambda_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm+1));

  for (t=0; t<(num_sm+1); t++) {  /* Be careful for the terminal condition */
    n = log_likelihood_sm_size[t];
    /* At maximum, we have n partial EM iterations and n full EM iterations: */
    log_likelihood_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
    log_prior_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
    lambda_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
  }
}

void alloc_log_itemp_sm(void)
{
  int n,t;

  log_itemp_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm+1));

  for (t=0; t<(num_sm+1); t++) {
    n = log_likelihood_sm_size[t];
    log_itemp_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
  }
}

void expand_log_likelihood_sm_all(int t)
{
  int old_log_likelihood_size;

  old_log_likelihood_size = log_likelihood_sm_size[t];
  log_likelihood_sm_size[t] *= 2;

  log_likelihood_sm[t] = (FLOAT_T *)n_realloc(log_likelihood_sm[t],
                                               sizeof(FLOAT_T)*log_likelihood_sm_size[t]);
  lambda_sm[t] = (FLOAT_T *)n_realloc(lambda_sm[t],
                                       sizeof(FLOAT_T)*log_likelihood_sm_size[t]);

  if (mode_smooth)
    log_prior_sm[t] = (FLOAT_T *)n_realloc(log_prior_sm[t],
                                            sizeof(FLOAT_T)*log_likelihood_sm_size[t]);

  if (mode_daem)
    log_itemp_sm[t] = (FLOAT_T *)n_realloc(log_itemp_sm[t],
                                            sizeof(FLOAT_T)*log_likelihood_sm_size[t]);

  log_likelihood_best = (FLOAT_T *)n_realloc(log_likelihood_best,sizeof(FLOAT_T)*log_likelihood_size);
  lambda_best = (FLOAT_T *)n_realloc(lambda_best,sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_smooth)
    log_prior_best = (FLOAT_T *)n_realloc(log_prior_best,sizeof(FLOAT_T)*log_likelihood_size);

  if (mode_daem) {
    log_itemp = (FLOAT_T *)n_realloc(log_itemp,sizeof(FLOAT_T)*log_likelihood_size);
    log_itemp_best = (FLOAT_T *)n_realloc(log_itemp_best,sizeof(FLOAT_T)*log_likelihood_size);
  }
}

void alloc_jmerge()
{
  int k;
  
  jmerge = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
  for (k=0; k<num_class; k++)
    jmerge[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
}

void alloc_jsplit()
{
  jsplit = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
}

void alloc_candidate()
{
  int h;
  
  sm_candidate = (int **)n_malloc(sizeof(int *)*num_candidate);
  for (h=0; h<num_candidate; h++)
    sm_candidate[h] = (int *)n_malloc(sizeof(int)*3);
}

void alloc_prob_part(void)
{
  int l;

  prob_incomp_part = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);

  prob_cond_part = (FLOAT_T **)n_malloc(sizeof(FLOAT_T*)*3);
  for (l=0; l<3; l++)
    prob_cond_part[l] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_entry);
}

void alloc_num_iterate_sm()
{
  num_iterate_sm_full = (int *)n_malloc(sizeof(int)*(num_sm+1));
  num_iterate_sm_part = (int *)n_malloc(sizeof(int)*(num_sm+1));
}

void alloc_num_result_sm()
{
  num_result_rejected = (int *)n_malloc(sizeof(int)*num_sm);
  num_result_skipped = (int *)n_malloc(sizeof(int)*num_sm);
}

void release_jmerge()
{
  int k;
  for (k=0; k<num_class; k++) {
    free(jmerge[k]); jmerge[k] = NULL;
  }
  free(jmerge); jmerge = NULL;
}

void release_jsplit()
{
  free(jsplit); jsplit = NULL;
}

void release_candidate()
{
  int h;

  for (h=0; h<num_candidate; h++) {
    free(sm_candidate[h]); sm_candidate[h] = NULL;
  }
  free(sm_candidate); sm_candidate = NULL;
}

void release_prob_part(void)
{
  int l;

  for (l=0; l<3; l++) {
    free(prob_cond_part[l]); prob_cond_part[l] = NULL;
  }
  free(prob_cond_part);   prob_cond_part = NULL;
  free(prob_incomp_part); prob_incomp_part = NULL;
}

void release_log_likelihood_sm_size(void)
{
  free(log_likelihood_sm_size);
  log_likelihood_sm_size = NULL;
}

void release_log_likelihood_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    free(log_likelihood_sm[t]); log_likelihood_sm[t] = NULL;
    free(log_prior_sm[t]); log_prior_sm[t] = NULL;
    free(lambda_sm[t]); lambda_sm[t] = NULL;
  }
  free(log_likelihood_sm); log_likelihood_sm = NULL;
  free(log_prior_sm); log_prior_sm = NULL;
  free(lambda_sm); lambda_sm = NULL;
}

void release_log_itemp_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    free(log_itemp_sm[t]); log_itemp_sm[t] = NULL;
  }
  free(log_itemp_sm); log_itemp_sm = NULL;
}

void release_num_iterate_sm()
{
  free(num_iterate_sm_full); num_iterate_sm_full = NULL;
  free(num_iterate_sm_part); num_iterate_sm_part = NULL;
}

void release_num_result_sm()
{
  free(num_result_rejected); num_result_rejected = NULL;
  free(num_result_skipped);  num_result_skipped = NULL;
}

void alloc_param_part(void)
{
  param_c_part = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*3);
}

void  release_param_part(void)
{
  free(param_c_part); param_c_part = NULL;
}

void alloc_param_sm(void)
{
  int k,v,u,a;

  param_c_sm = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    param_w_sm = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      param_w_sm[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	param_w_sm[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    param_u_sm = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    param_s_sm = (FLOAT_T ****)n_malloc(sizeof(FLOAT_T ***)*num_class);

    for (k=0; k<num_class; k++) {
      param_u_sm[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_size);
      param_s_sm[k] = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	param_u_sm[k][u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
	param_s_sm[k][u] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*gauss_attr_size[u]);

	for (a=0; a<gauss_attr_size[u]; a++)
	  param_s_sm[k][u][a] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_attr_size[u]);
      }
    }
  }
}

void release_param_sm(void)
{
  int k,v,u,a;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(param_w_sm[k][v]); param_w_sm[k][v] = NULL;
      }
      free(param_w_sm[k]); param_w_sm[k] = NULL;
    }
    free(param_w_sm); param_w_sm = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	free(param_u_sm[k][u]); 
	for (a=0; a<gauss_attr_size[u]; a++) {
	  free(param_s_sm[k][u][a]);
	}
      }
      free(param_u_sm[k]);
      free(param_s_sm[k]);
    }
    free(param_u_sm); param_u_sm = NULL;
    free(param_s_sm);  param_s_sm = NULL;
  }

  free(param_c_sm); param_c_sm = NULL;
}
