/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

void die(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_log,fmt,ap); fflush(fp_log);
  va_end(ap);

  va_start(ap,fmt);
  vfprintf(fp_msg,fmt,ap); fflush(fp_msg);
  va_end(ap);

  exit(1);
}

void die_s(int line_c, char *buf)
{
  die("Illegal line %d: %s",line_c,conv_msg(buf));
}

void die_line(int line_c, int id, char *buf)
{
  die("[ID:%d] Illegal line %d: %s",id,line_c,conv_msg(buf));
}

void die_line_long(char *fname, int line_c,
                   int id, char *buf, char *msg, int pos)
{
  char *ind;
  int i;

  ind = alloc_str(MAX_LINE_LEN + 1);

  if (pos >= MAX_LINE_LEN) {
    die("Internal error: die_line_long()\n");
  }
  else if (pos >= 0) {
    ind[pos] = '^';
    ind[pos+1] = '\0';
    for (i=0; i<pos; i++) ind[i] = ' ';
  }

  /* a new-line character is included in buf */
  if (pos < 0)
    die("[ID:%d] Illegal line %d in %s: %s\n%s",id,line_c,fname,msg,conv_msg(buf));
  else
    die("[ID:%d] Illegal line %d in %s: %s\n%s%s\n",id,line_c,fname,msg,conv_msg(buf),ind);
}

void die_alloc(char *fname, int line)
{
  fprintf(stderr,"Couldn't allocate memory (%s, line %d)\n",fname,line);
  exit(1);
}

void die_realloc(char *fname, int line)
{
  fprintf(stderr,"Couldn't reallocate memory (%s, line %d)\n",fname,line);
  exit(1);
}

void print_msg(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_msg,conv_msg(fmt),ap);
  va_end(ap);
  fflush(fp_msg);
}

void print_msg_matrix(MATRIX m,char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_msg,conv_msg(fmt),ap);
  va_end(ap);
  fprint_matrix(fp_msg,m);
  fflush(fp_msg);
}

void print_msg_vector(VECTOR v,char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_msg,conv_msg(fmt),ap);
  va_end(ap);
  fprint_vector(fp_msg,v);
  fflush(fp_msg);
}

/* CR/LF conversion and %.Nf -> %.{prec}f */
char *conv_base(char *prec_str, char *buf_in)
{
  char *bi, *bm;
  int l,i,state; /* state for %.Nf -> %.{prec}f */

  *buf_base = '\0';
  bi = buf_in;
  bm = buf_base;
  l = 0;
  state = 0;
#ifdef USE_CRLF
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_base): buffer overflow\n");
    if (*bi == '\n') {
      *bm = '\r';   /* insert CR */
      *(bm+1) = '\n';
      bm += 2;
      l += 2;
    }
    else if (*bi == '%') {
      state = 1;
      *bm = *bi;  bm++;  l++;
    }
    else if (state == 1 && *bi == '.') {
      state = 2;
      *bm = *bi;  bm++;  l++;
    }
    else if (state == 2 && *bi == 'N') {
      state = 3;
      if (l + strlen(prec_str) >= MAX_OUT_LEN)
        die("Internal error (conv_base): buffer overflow\n");
      for (i=0; i<strlen(prec_str); i++) {
        *bm = prec_str[i];
        bm++;
        l++;
      }
#ifndef NO_LDOUBLE
      *bm = 'L';  bm++;  l++;
#endif
    }
    else if (state == 3 && (*bi == 'e' || *bi == 'f' || *bi == 'g')) {
      state = 0;
      *bm = *bi;  bm++;  l++;
    }
    else {
      *bm = *bi;  bm++;  l++;
    }
    bi++;
  }
#else
  /* Only convert %.nnLf -> %.nnf */
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_base): buffer overflow\n");
    if (*bi == '%') {
      state = 1;
      *bm = *bi;  bm++;  l++;
    }
    else if (state == 1 && *bi == '.') {
      state = 2;
      *bm = *bi;  bm++;  l++;
    }
    else if (state == 2 && *bi == 'N') {
      state = 3;
      if (l + strlen(prec_str) >= MAX_OUT_LEN)
        die("Internal error (conv_base): buffer overflow\n");
      for (i=0; i<strlen(prec_str); i++) {
        *bm = prec_str[i];
        bm++;
        l++;
      }
#ifndef NO_LDOUBLE
      *bm = 'L';  bm++;  l++;
#endif
    }
    else if (state == 3 && (*bi == 'e' || *bi == 'f' || *bi == 'g')) {
      state = 0;
      *bm = *bi;  bm++;  l++;
    }
    else {
      *bm = *bi;  bm++;  l++;
    }
    bi++;
  }
#endif
  *bm = '\0';

  return buf_base;
}

/* Conversion from the internal charcode to the system's default charcode */
char *conv_sys(char *buf_in)
{
#if defined(USE_ICONV) && defined(CHARCODE)
  int buf_len_in0;
  size_t buf_len_in, buf_len_out;
  char *b;
#ifdef ICONV_CONST
  const char *bi;
#else
  char *bi;
#endif

  if (!mode_sys_out) return buf_in;

  *buf_sys = '\0';
  b = buf_sys;
  bi = buf_in;
  buf_len_in0 = strlen(buf_in);
  buf_len_in = (size_t)buf_len_in0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_sys_out,&bi,&buf_len_in,&b,&buf_len_out) == (size_t)-1)
    die("Couldn't convert: %s (errno=%d)\n",buf_in,errno);
  buf_sys[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_in;
#endif

  return buf_sys;
}

/* Alternative for the itoa() function
 * [Note] This function only returns a string allocated in the static area.
 *        So we should make a copy of the returned string if necessary.
 */
char *int2str(int i)
{
  static char s[INT2STR_BUFSIZE];
  int size;

  size = ((i > 0) ? (int)(LOG(i)/LOG(10)) + 1 : 1) + 1;
  if (size >= INT2STR_BUFSIZE)
    die("Internal error (int2str): buffer overflow\n");

#ifdef _MSC_VER
  _snprintf(s,(size - 1),"%d",i);
  s[size - 1] = '\0';
#else
  snprintf(s,size,"%d",i);
#endif

  return s;
}

char *int2strdup(int i)
{
  char *s0 = int2str(i);  /* s0 can be considered as null-terminated */
  char *s = alloc_str(strlen(s0) + 1);
  strcpy(s,s0);

  return s;
}

char *conv_msg(char *buf_in)
{
  static char *prec_msg_str = NULL;
  
  if (prec_msg_str == NULL) prec_msg_str = int2strdup(prec_msg);

  return conv_sys(conv_base(prec_msg_str,buf_in));
}

/* Conversion of the delimiters: ',' -> ',' or '\t' */
char *conv_delim(char *buf_in)
{
  char *bi,*bm;
  int l;

  *buf_delim = '\0';
  bi = buf_in;
  bm = buf_delim;
  l = 0;
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_delim): buffer overflow\n");
    if (*bi == '~') {
      bi++;
      if (*bi == '\0' || *bi == '\n')
        die("Internal error (conv_delim): illegal escape sequence\n");
      *bm = *bi;
      bm++; l++;
    }
    else if (*bi == ',') {
      *bm = (format_out == FMT_CSV) ? *bi : '\t';
      bm++; l++;
    }
    else {
      *bm = *bi;
      bm++; l++;
    }
    bi++;
  }
  *bm = '\0';

  return buf_delim;
}

/* Conversion from the internal charcode to the charcode specified by -q flag */
char *conv_out(char *buf_in)
{
#ifdef USE_ICONV
  int buf_len_in0;
  size_t buf_len_in, buf_len_out;
  char *b;
#  ifdef ICONV_CONST
  const char *bi;
#  else
  char *bi;
#  endif
#endif

#ifdef USE_ICONV
  if (!mode_conv_out) return buf_in;

  *buf_cout = '\0';
  b = buf_cout;
  bi = buf_in;
  buf_len_in0 = strlen(buf_in);
  buf_len_in = (size_t)buf_len_in0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_out,&bi,&buf_len_in,&b,&buf_len_out) == (size_t)-1)
    die("Couldn't convert: %s (errno=%d)\n",buf_in,errno);
  buf_cout[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_in;
#endif

  return buf_cout;
}

char *conv_desc(char *buf_in)
{
  static char *prec_out_str = NULL;

  if (prec_out_str == NULL)
    prec_out_str = int2strdup(prec_out);

  return conv_out(conv_base(prec_out_str,buf_in));
}

char *conv_data(char *buf_in)
{
  static char *prec_out_str = NULL;

  if (prec_out_str == NULL)
    prec_out_str = int2strdup(prec_out);

  return conv_out(conv_delim(conv_base(prec_out_str,buf_in)));
}

void conv_printf(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vprintf(conv_data(fmt),ap);
  va_end(ap);
}

void conv_fprintf(FILE *fp, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);
}

void conv_sprintf(char *s, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vsprintf(s,conv_data(fmt),ap);
  va_end(ap);
}

void conv_snprintf(char *s, size_t size, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(s,(size - 1),conv_data(fmt),ap);
  s[size - 1] = '\0';
#else
  vsnprintf(s,size,conv_data(fmt),ap);
#endif
  va_end(ap);
}

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)

/* get the current time in seconds */
FLOAT_T current_sec()
{
  struct rusage ru;
  FLOAT_T t;

  getrusage(RUSAGE_SELF, &ru);

  t =
    (FLOAT_T)(ru.ru_utime.tv_sec) +
    ((FLOAT_T)(ru.ru_utime.tv_usec))/1.0e+6;

  return t;
}

/* get the elasped time in seconds */
FLOAT_T elapsed_sec()
{
  struct timeval tv;
  FLOAT_T t;

  gettimeofday(&tv, NULL);

  t = (FLOAT_T)tv.tv_sec + ((FLOAT_T)tv.tv_usec)/1.0e+6;
  return t;
}
#endif

void print_dot_message(int m)
{
  if (m % 20 == 0) {
    if (m % 100 == 0) print_msg("%d",m);
    else print_msg(".");
    if (fp_msg == stderr) fflush(fp_msg);
  }
}

void print_daem_message(int m, FLOAT_T itemp)
{
  if (mode_daem) {
    if (m > 0 || itemp < 1.0) {
      /* print_msg("<%.Nf>",itemp); */
      print_msg("*");
    }
    if (fp_msg == stderr) fflush(fp_msg);
  }
}

void print_daem_message_sm(int t, int m, FLOAT_T itemp)
{
  if (mode_daem) {
    if (t == 0 && (m > 0 || itemp < 1.0)) {
      /* print_msg("<%.Nf>",itemp); */
      print_msg("*");
    }
    if (fp_msg == stderr) fflush(fp_msg);
  }
}

void toupper_str(char *s)
{
  int i;

  for (i=0; i<strlen(s); i++)
    s[i] = toupper(s[i]);
}

void tolower_str(char *s)
{
  int i;

  for (i=0; i<strlen(s); i++)
    s[i] = tolower(s[i]);
}

void *n_malloc_(size_t size, char *fname, int line)
{
  void *p;

  p = malloc(size);
  if (p==NULL) die_alloc(fname,line);

  return p;
}

void *n_realloc_(void *p, size_t size, char *fname, int line)
{
  p = realloc(p,size);
  if (p==NULL) die_realloc(fname,line);

  return p;
}

char *alloc_str_(int len, char *fname, int line)
{
  char *p;

  p = (char *)malloc(sizeof(char)*len);
  if (p==NULL) die_alloc(fname,line);

  memset(p,0,sizeof(char)*len);

  return p;
}

char *dup_str_(char *s, char *fname, int line)
{
  char *p;
  int len;

  len = strlen(s);
  p = (char *)malloc(sizeof(char)*(len + 1));
  if (p==NULL) die_alloc(fname,line);
  strcpy(p,s);

  return p;
}

int is_attr_id(int v)
{
  return (mode_id && v == attr_id) ? 1 : 0;
}

int is_attr_eval(int v)
{
  return (mode_eval && v == attr_eval) ? 1 : 0;
}

int is_attr_ignore(int v)
{
  return (mode_ignore && attr_ignore[v]) ? 1 : 0;
}

int is_attr_cont(int v)
{
  return attr_info[v].is_cont;
}

int is_attr_disc(int v)
{
  return (!(is_attr_id(v) || is_attr_eval(v) || is_attr_ignore(v) || is_attr_cont(v)));
}

#ifdef MAIN_MISC
#include "nbc_var.h"
int main(int argc, char *argv[])
{
  char **field;
  int f,num_field;
  int ret;
  char *buf;

  num_field = 4;
  buf = "fafa ,faff , f , a";

  ret = split_comma(num_field,buf,field);
  if (ret)
    fprintf(stderr,"Unexpected number of fields (#fields = %d):\n%s",num_field,buf);

  printf("num_field = %d\n",num_field);
  for (f=0; f<num_field; f++) {
    printf("%d: %s\n",f,field[f]);
  }

  return 0;
}
#endif
